/*
*	lr: list directories recursively
*	version: 2.0
*	date: 2017/04/14
*	auther: Kenar(Kenji Arisawa)
*	Email: arisawa@aichi-u.ac.jp
*
*	compiling under plan9port
*		mk -f mkfile_p9p
*/

#include <u.h>
#include <libc.h>
#include <bio.h>

#ifdef UNIX
#include <libgen.h>
#include <sys/stat.h>
#include <dirent.h>
/* I added some DMXXX. look libc.h in p9p */
#define DMBLOCK 0x00800000	/* same as DMDEVICE */
#define DMCHAR  0x00400000	/* we have room for new definition */
/* stime may conflict on some UNIXs */
#define stime s_time
#endif

#define DBG if(0)

void
usage(void)
{
	fprint(2,"usage: lr [-QqilnaL] [-d depth] [-D depth] [-t time]  $path ...\n"
	"lr lists directory recursively\n"
	"options:\n"
	"	-Q: unquoted format, instead TAB as separator (used only in combination with '-l')\n"
	"	-at time: show files accessed since the time (seconds since the unix epoch)\n"
	"	-d depth: search directory upto depth (depth >= 0)\n"
	"	-D depth: search directory of depth (depth >= 0)\n"
	"	-i: show date using iso-8601 style\n"
	"	-l: long format\n"
	"	-n: show date using unix epoch\n"
	"	-q: show qid\n"
	"	-t time: show files modified since the time (seconds since the unix epoch)\n"
	"	-L: follow links (for unix)\n"
	);
	exits("usage");
}

enum
{
	LFMT = 1,
	nFMT = 2,
	QFMT = 4,
	iFMT = 8,
	qFMT = 16,
};

char *mods[]={"---","--x","-w-","-wx","r--","r-x","rw-","rwx"};

int mode = 0;
Biobuf	out;
ulong stime = 0;
int aflag = 0; /* accessed time */
int Dflag = 0;
int Rflag = 0;
int Lflag = 0;

char*
modefmt(ulong m)
{
	static char buf[32];
	char *e, *s;
	s = buf;
	e = s + sizeof(buf);
#ifdef UNIX
	s[0] = '-';
	if(m & DMDIR)
		s[0] = 'd';
	if(m & DMNAMEDPIPE)
		s[0] = 'p';
	if(m & DMSOCKET)
		s[0] = 's';
	/* I have added DMBLOCK and DMCHAR */
	if(m & DMBLOCK)
		s[0] = 'b';
	if(m & DMCHAR)
		s[0] = 'c';
	if(!Lflag && (m & DMSYMLINK))
		s[0] = 'l';
	s += 1;
#else
	s[0] = s[1] = '-';
	if(m & DMDIR)
	s[0] = 'd';
	if(m & DMAPPEND)
		s[0] = 'a';
	if(m & DMEXCL)
		s[1] = 'l';
	s += 2;
#endif
	s = seprint(s,e, "%s", mods[(m & 0700) >> 6]);
	s = seprint(s,e, "%s", mods[(m & 070) >> 3]);
	seprint(s,e, "%s", mods[m & 07]);

#ifdef UNIX
	/* look $P9P/include/libc.h
	 * we have DMSETUID and DMSETGID
	 * however this is no effect. probably russ forgot implementation 
	 * look $P9P/src/lib9/_p9dir.c */
	if(m & DMSETUID)
		buf[3] = 's';
	if(m & DMSETGID)
		buf[6] = 's';
#endif
	return buf;
}

char *
datefmt(ulong t)
{
	static char buf[48];
	Tm *tm;
	char sep = '/';
	char fsep = ' ';

	if(mode & nFMT){
		snprint(buf,sizeof buf, "%lud", t);
		return buf;
	}
	tm = localtime(t);
	if(mode & iFMT)
		sep = '-';
	if(mode & QFMT)
		fsep = '\t';
	snprint(buf,sizeof buf,"%04d%c%02d%c%02d%c%02d:%02d:%02d",
		tm->year + 1900,sep,tm->mon + 1,sep,tm->mday, fsep,tm->hour, tm->min, tm->sec);
	return buf;
}

char *
qidfmt(Qid qid)
{
	static char buf[48];
	if(mode & QFMT)
		snprint(buf,sizeof buf, "%.16llux\t%.5lud\t%.2ux", qid.path, qid.vers, qid.type);
	else
		snprint(buf,sizeof buf, "%.16llux %.5lud %.2ux", qid.path, qid.vers, qid.type);
	return buf;
}


/*
*	don't use d->name, the name is in path
*/
void
ls(char *path, Dir *d)
{
	char *fmt;
	char *sep;

	if(mode & qFMT){
		if(mode & LFMT){
			fmt = "%s %s %-8s %-8s %10llud %s %q\n";
			if(mode & QFMT)
				fmt = "%s\t%s\t%s\t%s\t%llud\t%s\t%s\n";
			Bprint(&out,fmt, qidfmt(d->qid), modefmt(d->mode),
			d->uid, d->gid, d->length, datefmt(d->mtime), path);
		}
		else{
			sep = " ";
			if(mode & QFMT)
				sep = "\t";
			fmt  = "%s%s%s\n";
			if(d->mode & DMDIR)
				fmt = "%s%s%s/\n";
			Bprint(&out,fmt, qidfmt(d->qid), sep, path);
		}
	}
	else
	if(mode & LFMT){
		fmt = "%s %-8s %-8s %10llud %s %q\n";
		if(mode & QFMT)
			fmt = "%s\t%s\t%s\t%llud\t%s\t%s\n";
		Bprint(&out,fmt, modefmt(d->mode),
		d->uid, d->gid, d->length, datefmt(d->mtime), path);
	}
	else{
		fmt  = "%s\n";
		if(d->mode & DMDIR)
			fmt = "%s/\n";
		Bprint(&out,fmt, path);
	}
}

int
compar(Dir *a, Dir *b)
{
	return strcmp(a->name, b->name);
}


#ifdef UNIX

typedef struct List List;
struct List{
	vlong	*qpath;
	List	*next;
};

List *pathlist;

/* chkpath append path to pl if absent
 * if already exist it returns -1 */
int
chkpath(List **pl, char *qidpath, int flag)
{
	List *p, *q;
	q = nil;
	p = *pl;
	while(p){
		if(p->qpath == qidpath)
			return -1;
		q = p;
		p = p->next;
	}
	/* then p is nil */
	if(!flag)
		return 0;
	p = malloc(sizeof(List));
	if(q)
		q->next = p;
	else
		*pl = p;
	p->qpath = qidpath;
	p->next = nil;
	return 0;
}

#endif

void
list(char *path, char *ep, int depth)
{
	Dir *db,*db0;
	char *p;
	int fd, i,n;
	ulong m;

#ifdef UNIX
	struct stat sb;

#endif

	/* NB: this dirstat() is from russ
	 * therefore not be reflected my change in _mp9dir.c */
	db = dirstat(path);
	if(db == nil){
		Bflush(&out);
		fprint(2, "# dirstat %s: %r\n", path);	
		return;
	}
	m = db->mode;

#ifdef UNIX
	/* use _mp9dir.c instead of _p9dir.c */
	if(stat(path, &sb) < 0){
		/* stat() is a unix function.
		 * it seems stat() detects cyclic link,
		 * not only broken link */
		Bflush(&out);
		fprint(2,"# %s: broken or cyclic link\n", path);
		free(db);
		return;
	}

	if(m & DMSYMLINK){
		if(Lflag && (m & DMDIR)){
			/* then the path is a link to a directory.
			 * we are considering:
			 *   path = base/ln
			 * ln is a link to a dir
			 * strategy for safe recursion:
			 *   put a mark to base
			 *   test if the dir is already marked or not */

			p = dirname(path);/* unix function */
			db0 = dirstat(p);
			/* free(p); don't free. not malloc'ed */

			/* put a check mark to the base dir */
			chkpath(&pathlist,db0->qid.path,1);

			/* examine: check if the dir (linked to) is already marked */
			if(chkpath(&pathlist,db->qid.path,0) < 0){
				Bflush(&out);
				fprint(2,"# %s: cyclic directory link\n", path);
				return;
			}
			free(db0);
		}
	}
#endif

	if(db->mtime > stime || (aflag && db->atime > stime)){
		if(!Dflag || (Dflag && (depth == 0)))
			ls(path, db);
	}
	free(db);

#ifdef UNIX
	if(!Lflag && (m & DMSYMLINK))
		return;
#endif

	/* we should return here if the path is file */
	if((m & DMDIR) == 0)
		return;

#ifdef UNIX
	if(Rflag && Lflag){
		/* put a check mark to this dir */
		if(chkpath(&pathlist,db->qid.path,1) < 0){
			Bflush(&out);
			fprint("# %s: cyclic directory link\n", path);
			return;
		}
	}
#endif

	if(depth == 0)
		return;
	depth--;

	fd = open(path, OREAD);
	if(fd < 0){
		Bflush(&out);
		fprint(2, "# %s: unreadable\n", path);
		return;
	}

	n = dirreadall(fd, &db);
	close(fd);
	if(n < 0){
		/* I believe db should not be allocated in case of error,
		 * so free(db) is not here */
		Bflush(&out);
		fprint(2, "# dirreadall %s: %r\n", path);
		return;
	}

	/* sort by name */
	qsort(db, n, sizeof(Dir), (int (*)(const void*, const void*))compar);// for unix

	for(i = 0; i < n; i++){
		if(db[i].name[0] == 0){
			/*  this case should not happen.
			*	but old cwfs (before 2013) allowed names that break UTF-8
			*	therefore we need some checks	-Kenar- */
			continue;
		}
		if(db[i].atime <= stime)
			continue;
		m = strlen(path);
		if(path[m-1] == '/')
			seprint(path + m - 1,ep,"/%s", db[i].name);
		else
			seprint(path + m,ep,"/%s", db[i].name);
		list(path, ep, depth);
		path[m] = 0;
	}
	free(db);
	return;
}


void main(int argc, char *argv[])
{
	char path[4096];
	char *ep, *ap,*p;
	int depth = -1;
	ARGBEGIN{
	case 'a':	aflag = 1; break;
	case 'D':	Dflag++; /* continue next case */
	case 'd':
		ap = EARGF(usage());
		depth = atoi(ap);
		break;
	case 'l':	mode |= LFMT; break;
	case 'n':	mode |= nFMT; break;
	case 'Q':	mode |= QFMT; break;
	case 'i':	mode |= iFMT; break;
	case 'q':	mode |= qFMT; break;
	case 't':
		ap = EARGF(usage());
		stime = (ulong)atol(ap);
		break;
	case 'R':	Rflag = 1; break; /* to be removed */
	case 'L':	Lflag = 1; break;
	default: usage();
	}ARGEND

	quotefmtinstall();
	Binit(&out, 1, OWRITE);
	ep = path + sizeof path;
	if(*argv == nil)
		usage();
	while(*argv){
		strecpy(path, ep, *argv);
		p = path;
		if(strcmp(path,"/") == 0)
			sysfatal("path \"/\" is not allowed");
		if(path[strlen(p)-1] == '/')
			path[strlen(p)-1] = 0;
		if(p[0] == '.' && p[1] == '/')
			p += 2;
		list(p, ep, depth);
		argv++;
	}
	exits(nil);
}
